#include "RAW.h"

#include <stdio.h>

#define PCMCHUNKSIZE    32760
#define SAMPLERATE      44100
#define BITRESOLUTION   16
#define NUMCHANNELS     2

extern "C" __declspec(dllexport) CInputPlugin *Apollo_GetInputModule2(HINSTANCE hDLLInstance,HWND hApolloWnd)
{
    return new CRAWPlugin(hDLLInstance,hApolloWnd);
}

char *CRAWPlugin::GetDescription()
{
    return "Illustrative raw PCM decoder";
}

CInputDecoder *CRAWPlugin::Open(char *filename,int audioDataOffset)
{
    // returning NULL means trouble initializing decoder

    return new CRAWDecoder(filename,audioDataOffset);
}

BOOL CRAWPlugin::GetInfo(char *filename,int audioDataOffset,TrackInfo *trackInfo)
{
    if (strstr(filename,"://") != NULL) return FALSE;   // we can't handle URLs

    HANDLE hFile = CreateFile(filename,GENERIC_READ,FILE_SHARE_READ | FILE_SHARE_WRITE,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
    
    if (hFile == INVALID_HANDLE_VALUE) return FALSE;    // error opening file

    trackInfo->fileSize = GetFileSize(hFile,NULL);

    CloseHandle(hFile);

    trackInfo->seekable = TRUE;
    trackInfo->hasEqualizer = FALSE;        // we want the player to provide equalization
    trackInfo->sampleRate = SAMPLERATE;
    trackInfo->bitResolution = BITRESOLUTION;
    trackInfo->numChannels = NUMCHANNELS;
    trackInfo->playingTime = (int)((__int64)(trackInfo->fileSize - audioDataOffset) * 1000 / (SAMPLERATE * NUMCHANNELS * (BITRESOLUTION >> 3)));
    trackInfo->bitRate = SAMPLERATE * NUMCHANNELS * BITRESOLUTION;
        // if the bitrate is not constant, store the average value with a negative sign

    strcpy(trackInfo->fileTypeDescription,"Raw PCM File");

    // if the file contains information about the title you should
    // use it as the suggested title instead of the filebody

    char *extensionBegin = strrchr(filename,'.');

    if (extensionBegin == NULL)
        extensionBegin = filename + strlen(filename);

    char *pathEnd = strrchr(filename,'\\');

    if (pathEnd == NULL)
    {
        memcpy(trackInfo->suggestedTitle,filename,extensionBegin - filename);
        trackInfo->suggestedTitle[extensionBegin - filename] = 0;
    }
    else
    {
        memcpy(trackInfo->suggestedTitle,pathEnd + 1,extensionBegin - pathEnd - 1);
        trackInfo->suggestedTitle[extensionBegin - pathEnd - 1] = 0;
    }

    return TRUE;
}

void CRAWPlugin::AdditionalInfo(char *filename,int audioDataOffset)
{
    // opens a dialog/window containing extra information about the file

    MessageBox(GetActiveWindow(),"No additional info","Raw Power!",MB_OK);
}

BOOL CRAWPlugin::IsSuitableFile(char *filename,int audioDataOffset)
{
    // here you should detect whether this plug-in supports the file or not

    if (strstr(filename,"://") != NULL) return FALSE;   // we can handle URLs

    char *extensionBegin = strrchr(filename,'.');

    if (extensionBegin == NULL) return FALSE;

    if (!stricmp(extensionBegin,".raw")) return TRUE;
    if (!stricmp(extensionBegin,".pcm")) return TRUE;

    return FALSE;
}

char *CRAWPlugin::GetExtensions()
{
    // returns a string containing all supported file extensions + their descriptions
    // format: "description1\0*.ext1;*.ext2;*.ext3\0description2\0*.ext1;*.ext2;*.ext3\0\0"
    // this information will be used in open/add dialogs and with shell integration

    return "Raw PCM data\0*.RAW;*.PCM\0\0";
}

void CRAWPlugin::Config()
{
    // opens a configuration dialog
    // the plug-in may store its configuration wherever it desires

    MessageBox(GetActiveWindow(),"Nothing to configurate","Raw Power!",MB_OK);
}

void CRAWPlugin::About()
{
    // opens an about dialog

    MessageBox(GetActiveWindow(),"Illustrative raw PCM input plug-in by Heikki Ylinen","Raw Power!",MB_OK);
}


CRAWDecoder::CRAWDecoder(char *filename,int audioDataOffset)
{
    _buffer = NULL;
    _beginOfRAWData = audioDataOffset;
    _hFile = INVALID_HANDLE_VALUE;

    if (strstr(filename,"://") != NULL) return;         // we can handle URLs

    _hFile = CreateFile(filename,GENERIC_READ,FILE_SHARE_READ | FILE_SHARE_WRITE,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);
    
    if (_hFile == INVALID_HANDLE_VALUE) return;          // error opening file

    _playingTime = (int)((__int64)(GetFileSize(_hFile,NULL) - _beginOfRAWData) * 1000 / (SAMPLERATE * NUMCHANNELS * (BITRESOLUTION >> 3)));
    _buffer = new char[PCMCHUNKSIZE];

    // seek to the beginning of the actual data

    SetFilePointer(_hFile,_beginOfRAWData,NULL,FILE_BEGIN);
}

CRAWDecoder::~CRAWDecoder()
{
    if (_hFile != INVALID_HANDLE_VALUE) CloseHandle(_hFile);
    if (_buffer != NULL) delete [] _buffer;
}

int CRAWDecoder::GetNextAudioChunk(void **buffer)
{
    if (_hFile == INVALID_HANDLE_VALUE) return 0;
        // invalid file, return 0 to indicate end of data

    DWORD numBytes;

    ReadFile(_hFile,_buffer,PCMCHUNKSIZE,&numBytes,NULL);

    // set the buffer-parameter to point to our data
    // the player will not manipulate this data

    *buffer = _buffer;

    // return the data size in bytes
    // the size of decoded data may vary from call to call,
    // but it must not be over 50K and it must be
	// aligned by the sample size
    
    // returning 0 means end of data
    // returning -1 means error in reading data

    return numBytes;
}

int CRAWDecoder::SetPosition(int newPosition)
{
    // the position is given in milliseconds

    __int64 position = newPosition;

    if (position > _playingTime) position = _playingTime;
    if (position < 0) position = 0;

    // round the seek position appropriately

    position *= SAMPLERATE * NUMCHANNELS * (BITRESOLUTION >> 3);
    position /= 1000;
    position /= NUMCHANNELS * (BITRESOLUTION >> 3);
    position *= NUMCHANNELS * (BITRESOLUTION >> 3);

    SetFilePointer(_hFile,_beginOfRAWData + (int)position,NULL,FILE_BEGIN);

    // calculate the actual seek point in milliseconds and return it

    position *= 1000;
    position /= SAMPLERATE * NUMCHANNELS * (BITRESOLUTION >> 3);

    return (int)position;
}

void CRAWDecoder::AdjustEqualizer(int equalizerValues[16])
{
    // the equlizer values are in decibels
    // their respective frequencies in Hz are
    // 50,130,220,300,430,600,820,1160,1680,2370,3320,4690,6680,9430,13300,18800

    // we don't need to do anything here since we set trackInfo->hasEqualizer = FALSE
}

void CRAWDecoder::SetEqualizer(BOOL equalize)
{
    // turns the equalizer on/off

    // we don't need to do anything here since we set trackInfo->hasEqualizer = FALSE
}
